@@ -2,23 +2,20 @@
 pub mod errors;
 mod jobs;
 mod man;
-mod quote;
+mod redirection;
 
 use std::env;
-use std::fs;
-use std::io::{self, BufRead, BufReader};
+use std::fs::{self, File};
+use std::io::{self, BufRead, BufReader, BufWriter, Write};
 use std::num::ParseIntError;
-use std::path::Path;
+use std::path::{Path, PathBuf};
 use std::process::exit;
 use std::time::Duration;
 
 use arrayvec::ArrayVec;
 use permutate::Permutator;
+use tokenizer::Token;
 use num_cpus;
-
-use super::disk_buffer::{self, DiskBufferTrait, DiskBufferWriter};
-use super::input_iterator::InputIterator;
-use super::tokenizer::Token;
 use self::errors::ParseErr;
 
 // Re-export key items from internal modules.
@@ -34,25 +31,25 @@ pub const QUIET_MODE:          u16 = 8;
 pub const VERBOSE_MODE:        u16 = 16;
 pub const DASH_EXISTS:         u16 = 32;
 pub const DRY_RUN:             u16 = 64;
-pub const ETA:                 u16 = 128;
-
-/// Defines what quoting mode to use when expanding the command.
-enum Quoting { None, Basic, Shell }
+pub const SHELL_QUOTE:         u16 = 128;
+pub const ETA:                 u16 = 256;
+pub const JOBLOG:              u16 = 512;
 
 /// `Args` is a collection of critical options and arguments that were collected at
 /// startup of the application.
-pub struct Args<'a> {
+pub struct Args {
     pub flags:     u16,
     pub ncores:    usize,
     pub ninputs:   usize,
     pub memory:    u64,
     pub delay:     Duration,
     pub timeout:   Duration,
-    pub arguments: ArrayVec<[Token<'a>; 128]>,
+    pub arguments: ArrayVec<[Token; 128]>,
+    pub joblog:    Option<String>,
 }
 
-impl<'a> Args<'a> {
-    pub fn new() -> Args<'a> {
+impl Args {
+    pub fn new() -> Args {
         Args {
             ncores:    num_cpus::get(),
             flags:     0,
@@ -61,34 +58,46 @@ impl<'a> Args<'a> {
             memory:    0,
             delay:     Duration::from_millis(0),
             timeout:   Duration::from_millis(0),
+            joblog:    None,
         }
     }
 
-    pub fn parse(&mut self, comm: &mut String, arguments: &[String], unprocessed_path: &Path) -> Result<InputIterator, ParseErr> {
+    /// Performs all the work related to parsing program arguments
+    pub fn parse(&mut self, comm: &mut String, arguments: &[String], unprocessed_path: &Path)
+        -> Result<usize, ParseErr>
+    {
         // Create a write buffer that automatically writes data to the disk when the buffer is full.
-        let mut disk_buffer = disk_buffer::DiskBuffer::new(unprocessed_path).write()
+        let disk_buffer = fs::OpenOptions::new().write(true).append(true).open(unprocessed_path)
             .map_err(|why| ParseErr::File(FileErr::Open(unprocessed_path.to_owned(), why)))?;
+        let mut disk_buffer = BufWriter::new(disk_buffer);
 
-        // Temporary stores for input arguments.
+        // Each list will consist of a series of input arguments
         let mut lists: Vec<Vec<String>>     = Vec::new();
+        // The `current_inputs` variable will contain all the inputs that have been collected for the first list.
         let mut current_inputs: Vec<String> = Vec::with_capacity(1024);
-        let mut number_of_arguments = 0;
+        // If this value is set, input arguments will be grouped into pairs defined by `max_args` value.
         let mut max_args = 0;
-        let mut quote = Quoting::None;
+        // It is important for the custom `InputIterator` to know how many input arguments are to be processed.
+        let mut number_of_arguments = 0;
 
+        // If no arguments were passed, we can assume that the standard input will be parsing commands.
+        // Otherwise, we will parse all the arguments and take actions based on these inputs.
         if env::args().len() > 1 {
-            // The purpose of this is to set the initial parsing mode.
+            // The first argument defines which `mode` to shift into and which argument `index` to start from.
             let (mut mode, mut index) = match arguments[1].as_str() {
-                ":::"  => { self.flags |= INPUTS_ARE_COMMANDS; (Mode::Inputs, 2) },
-                "::::" => { self.flags |= INPUTS_ARE_COMMANDS; (Mode::Files, 2) },
-                _  => (Mode::Arguments, 1)
+                ":::"  | ":::+"  => (Mode::Inputs, 2),
+                "::::" | "::::+" => (Mode::Files, 2),
+                _                => (Mode::Arguments, 1)
             };
 
+            // If the `--shebang` parameter was passed, this will be set to `true`.
             let mut shebang = false;
 
             if let Mode::Arguments = mode {
+                // Parse arguments until the command has been found.
                 while let Some(argument) = arguments.get(index) {
                     index += 1;
+
                     let mut char_iter = argument.chars();
 
                     // If the first character is a '-' then it will be processed as an argument.
@@ -108,7 +117,6 @@ impl<'a> Args<'a> {
                                         exit(0);
                                     },
                                     'p' => self.flags |= PIPE_IS_ENABLED,
-                                    'q' => quote = Quoting::Basic,
                                     's' => self.flags |= QUIET_MODE,
                                     'v' => self.flags |= VERBOSE_MODE,
                                     _ => {
@@ -131,6 +139,12 @@ impl<'a> Args<'a> {
                                     println!("{}", man::MAN_PAGE);
                                     exit(0);
                                 },
+                                "joblog" => {
+                                    let file = arguments.get(index).ok_or(ParseErr::JoblogNoValue)?;
+                                    self.joblog = Some(file.to_owned());
+                                    index += 1;
+                                    self.flags |= JOBLOG;
+                                },
                                 "jobs" => {
                                     let val = arguments.get(index).ok_or(ParseErr::JobsNoValue)?;
                                     self.ncores = jobs::parse(val)?;
@@ -152,8 +166,7 @@ impl<'a> Args<'a> {
                                 }
                                 "pipe" => self.flags |= PIPE_IS_ENABLED,
                                 "quiet" | "silent" => self.flags |= QUIET_MODE,
-                                "quote" => quote = Quoting::Basic,
-                                "shellquote" => quote = Quoting::Shell,
+                                "shellquote" => self.flags |= DRY_RUN + SHELL_QUOTE,
                                 "timeout" => {
                                     let val = arguments.get(index).ok_or(ParseErr::TimeoutNoValue)?;
                                     let seconds = val.parse::<f64>().map_err(|_| ParseErr::TimeoutNaN(index))?;
@@ -162,19 +175,7 @@ impl<'a> Args<'a> {
                                 }
                                 "verbose" => self.flags |= VERBOSE_MODE,
                                 "version" => {
-                                    println!("parallel 0.9.0\n\nCrate Dependencies:");
-                                    println!("    arrayvec     0.3.20");
-                                    println!("    gcc          0.3.41");
-                                    println!("    kernel32-sys 0.2.2");
-                                    println!("    libc         0.2.18");
-                                    println!("    num_cpus     1.2.1");
-                                    println!("    nodrop       0.1.8");
-                                    println!("    odds         0.2.25");
-                                    println!("    permutate    0.2.0");
-                                    println!("    sys-info     0.4.1");
-                                    println!("    wait-timeout 0.1.3");
-                                    println!("    winapi       0.2.8");
-                                    println!("    winapi-build 0.1.1");
+                                    println!("MIT/Rust Parallel 0.10.0\n");
                                     exit(0);
                                 }
                                 _ if &argument[2..9] == "shebang" => {
@@ -187,14 +188,8 @@ impl<'a> Args<'a> {
                         }
                     } else {
                         match argument.as_str() {
-                            ":::" => {
-                                mode = Mode::Inputs;
-                                self.flags |= INPUTS_ARE_COMMANDS;
-                            },
-                            "::::" => {
-                                mode = Mode::Files;
-                                self.flags |= INPUTS_ARE_COMMANDS;
-                            }
+                            ":::" => mode = Mode::Inputs,
+                            "::::" => mode = Mode::Files,
                             _ => {
                                 // The command has been supplied, and argument parsing is over.
                                 comm.push_str(argument);
@@ -206,7 +201,9 @@ impl<'a> Args<'a> {
                 }
             }
 
-            if let Mode::Command = mode {
+            if let Some(path) = redirection::input_was_redirected() {
+                file_parse(&mut current_inputs, path.to_str().ok_or_else(|| ParseErr::RedirFile(path.clone()))?)?;
+            } else if let Mode::Command = mode {
                 while let Some(argument) = arguments.get(index) {
                     index += 1;
                     match argument.as_str() {
@@ -223,48 +220,50 @@ impl<'a> Args<'a> {
                     }
                     break
                 }
-            }
 
-            if shebang {
-                file_parse(&mut current_inputs, &arguments.last().unwrap())?;
+                if shebang {
+                    file_parse(&mut current_inputs, &arguments.last().unwrap())?;
+                } else {
+                    parse_inputs(arguments, index, &mut current_inputs, &mut lists, &mut mode)?;
+                }
             } else {
                 parse_inputs(arguments, index, &mut current_inputs, &mut lists, &mut mode)?;
             }
 
-            number_of_arguments = write_inputs_to_disk(lists, current_inputs, max_args, &mut disk_buffer)?;
+            number_of_arguments = write_inputs_to_disk(lists, current_inputs, max_args, &mut disk_buffer, unprocessed_path)?;
+        } else if let Some(path) = redirection::input_was_redirected() {
+            file_parse(&mut current_inputs, path.to_str().ok_or_else(|| ParseErr::RedirFile(path.clone()))?)?;
+            number_of_arguments = write_inputs_to_disk(lists, current_inputs, max_args, &mut disk_buffer, unprocessed_path)?;
         }
 
-        if disk_buffer.is_empty() {
-            number_of_arguments = write_stdin_to_disk(&mut disk_buffer, max_args)?;
+        if number_of_arguments == 0 {
+            number_of_arguments = write_stdin_to_disk(&mut disk_buffer, max_args, unprocessed_path)?;
         }
 
         if number_of_arguments == 0 { return Err(ParseErr::NoArguments); }
 
         // Flush the contents of the buffer to the disk before tokenizing the command argument.
-        disk_buffer.flush().map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+        let _ = disk_buffer.flush();
 
-        // Expand the command if quoting is enabled
-        match quote {
-            Quoting::None  => (),
-            Quoting::Basic => *comm = quote::basic(comm.as_str()),
-            Quoting::Shell => *comm = quote::shell(comm.as_str()),
-        }
+        if comm.is_empty() { self.flags |= INPUTS_ARE_COMMANDS; }
 
-        // Return an `InputIterator` of the arguments contained within the unprocessed file.
-        let inputs = InputIterator::new(unprocessed_path, number_of_arguments).map_err(ParseErr::File)?;
-        Ok(inputs)
+        Ok(number_of_arguments)
     }
 }
 
-fn write_stdin_to_disk(disk_buffer: &mut DiskBufferWriter, max_args: usize) -> Result<usize, ParseErr> {
+/// Write all arguments from standard input to the disk, recording the number of arguments that were read.
+fn write_stdin_to_disk(disk_buffer: &mut BufWriter<File>, max_args: usize, unprocessed_path: &Path)
+    -> Result<usize, ParseErr>
+{
+    println!("parallel: reading inputs from standard input");
     let mut number_of_arguments = 0;
 
     let stdin = io::stdin();
     if max_args < 2 {
         for line in stdin.lock().lines() {
             if let Ok(line) = line {
-                disk_buffer.write(line.as_bytes()).and_then(|_| disk_buffer.write_byte(b'\n'))
-                    .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                disk_buffer.write(line.as_bytes()).and_then(|_| disk_buffer.write(b"\n"))
+                    .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 number_of_arguments += 1;
             }
         }
@@ -276,32 +275,33 @@ fn write_stdin_to_disk(disk_buffer: &mut DiskBufferWriter, max_args: usize) -> R
                     max_args_index -= 1;
                     number_of_arguments += 1;
                     disk_buffer.write(line.as_bytes())
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 } else if max_args_index == 1 {
                     max_args_index = max_args;
-                    disk_buffer.write_byte(b' ')
+                    disk_buffer.write(b" ")
                         .and_then(|_| disk_buffer.write(line.as_bytes()))
-                        .and_then(|_| disk_buffer.write_byte(b'\n'))
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .and_then(|_| disk_buffer.write(b"\n"))
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 } else {
                     max_args_index -= 1;
-                    disk_buffer.write_byte(b' ')
+                    disk_buffer.write(b" ")
                         .and_then(|_| disk_buffer.write(line.as_bytes()))
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 }
             }
         }
         if max_args_index != max_args {
-            disk_buffer.write_byte(b'\n')
-                .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+            disk_buffer.write(b"\n")
+                .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
         }
     }
 
     Ok(number_of_arguments)
 }
 
+/// Write all input arguments buffered in memory to the disk, recording the number of arguments that were read.
 fn write_inputs_to_disk(lists: Vec<Vec<String>>, current_inputs: Vec<String>, max_args: usize,
-    disk_buffer: &mut DiskBufferWriter) -> Result<usize, ParseErr> {
+    disk_buffer: &mut BufWriter<File>, unprocessed_path: &Path) -> Result<usize, ParseErr> {
     let mut number_of_arguments = 0;
 
     if lists.len() > 1 {
@@ -321,28 +321,27 @@ fn write_inputs_to_disk(lists: Vec<Vec<String>>, current_inputs: Vec<String>, ma
         {
             let mut iter = permutation_buffer.iter();
             disk_buffer.write(iter.next().unwrap().as_bytes())
-                .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
             for element in iter {
-                disk_buffer.write_byte(b' ').and_then(|_| disk_buffer.write(element.as_bytes()))
-                    .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                disk_buffer.write(b" ").and_then(|_| disk_buffer.write(element.as_bytes()))
+                    .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
             }
 
             number_of_arguments += 1;
         }
 
-        // Reuse that buffer for each successive permutation
         if max_args < 2 {
-            disk_buffer.write_byte(b'\n').map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+            disk_buffer.write(b"\n").map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
             while let Ok(true) = permutator.next_with_buffer(&mut permutation_buffer) {
                 let mut iter = permutation_buffer.iter();
                 disk_buffer.write(iter.next().unwrap().as_bytes())
-                    .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                    .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 for element in iter {
-                    disk_buffer.write_byte(b' ').and_then(|_| disk_buffer.write(element.as_bytes()))
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                    disk_buffer.write(b" ").and_then(|_| disk_buffer.write(element.as_bytes()))
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 }
-                disk_buffer.write_byte(b'\n')
-                    .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                disk_buffer.write(b"\n")
+                    .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 number_of_arguments += 1;
             }
         } else {
@@ -354,34 +353,34 @@ fn write_inputs_to_disk(lists: Vec<Vec<String>>, current_inputs: Vec<String>, ma
                     number_of_arguments += 1;
 
                     disk_buffer.write(iter.next().unwrap().as_bytes())
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
 
                     for element in iter {
-                        disk_buffer.write_byte(b' ').and_then(|_| disk_buffer.write(element.as_bytes()))
-                            .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        disk_buffer.write(b" ").and_then(|_| disk_buffer.write(element.as_bytes()))
+                            .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                     }
                 } else if max_args_index == 1 {
                     max_args_index = max_args;
-                    disk_buffer.write_byte(b' ')
+                    disk_buffer.write(b" ")
                         .and_then(|_| disk_buffer.write(iter.next().unwrap().as_bytes()))
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
 
                     for element in iter {
-                        disk_buffer.write_byte(b' ').and_then(|_| disk_buffer.write(element.as_bytes()))
-                            .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        disk_buffer.write(b" ").and_then(|_| disk_buffer.write(element.as_bytes()))
+                            .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                     }
 
-                    disk_buffer.write_byte(b'\n')
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                    disk_buffer.write(b"\n")
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 } else {
                     max_args_index -= 1;
-                    disk_buffer.write_byte(b' ')
+                    disk_buffer.write(b" ")
                         .and_then(|_| disk_buffer.write(iter.next().unwrap().as_bytes()))
-                        .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
 
                     for element in iter {
-                        disk_buffer.write_byte(b' ').and_then(|_| disk_buffer.write(element.as_bytes()))
-                            .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                        disk_buffer.write(b" ").and_then(|_| disk_buffer.write(element.as_bytes()))
+                            .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                     }
                 }
             }
@@ -389,8 +388,8 @@ fn write_inputs_to_disk(lists: Vec<Vec<String>>, current_inputs: Vec<String>, ma
     } else if max_args < 2 {
         for input in current_inputs {
             disk_buffer.write(input.as_bytes())
-                .and_then(|_| disk_buffer.write_byte(b'\n'))
-                .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                .and_then(|_| disk_buffer.write(b"\n"))
+                .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
             number_of_arguments += 1;
         }
     } else {
@@ -401,20 +400,22 @@ fn write_inputs_to_disk(lists: Vec<Vec<String>>, current_inputs: Vec<String>, ma
 
             while index != max_index {
                 disk_buffer.write(chunk[index].as_bytes())
-                    .and_then(|_| disk_buffer.write_byte(b' '))
-                    .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                    .and_then(|_| disk_buffer.write(b" "))
+                    .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
                 index += 1;
             }
             disk_buffer.write(chunk[max_index].as_bytes())
-                .and_then(|_| disk_buffer.write_byte(b'\n'))
-                .map_err(|why| FileErr::Write(disk_buffer.path.clone(), why))?;
+                .and_then(|_| disk_buffer.write(b"\n"))
+                .map_err(|why| FileErr::Write(PathBuf::from(unprocessed_path), why))?;
         }
     }
     Ok(number_of_arguments)
 }
 
-fn parse_inputs(arguments: &[String], mut index: usize, current_inputs: &mut Vec<String>, lists: &mut Vec<Vec<String>>,
-    mode: &mut Mode) -> Result<(), ParseErr> {
+/// Collects all the provided inputs that were passed as command line arguments into the program.
+fn parse_inputs(arguments: &[String], mut index: usize, current_inputs: &mut Vec<String>,
+    lists: &mut Vec<Vec<String>>, mode: &mut Mode) -> Result<(), ParseErr>
+{
     let mut append_list = &mut Vec::new();
 
     macro_rules! switch_mode {
@@ -498,6 +499,8 @@ fn merge_lists(original: &mut Vec<String>, append: &mut Vec<String>) {
     }
 }
 
+/// When the `--memfree` option has been selected, this will attempt to parse the unit's value, multiplying
+/// that value by the unit's multiplier.
 fn parse_memory(input: &str) -> Result<u64, ParseIntError> {
     let result = match input.chars().last().unwrap() {
         'k' => &input[..input.len()-1].parse::<u64>()? * 1_000,
