@@ -1,22 +1,24 @@
 #![deny(dead_code)]
-#![deny(unused_imports)]
 #![allow(unknown_lints)]
 #![feature(loop_break_value)]
 #![feature(alloc_system)]
 extern crate alloc_system;
 extern crate arrayvec;
+extern crate itoa;
 extern crate num_cpus;
 extern crate permutate;
+extern crate smallvec;
 extern crate sys_info;
+extern crate time;
 extern crate wait_timeout;
 
 mod arguments;
-mod command;
 mod disk_buffer;
 mod execute;
 mod filepaths;
 mod init;
 mod input_iterator;
+mod misc;
 mod tokenizer;
 mod shell;
 mod verbose;
@@ -27,22 +29,28 @@ use std::io::{self, BufRead, BufReader, Write};
 use std::mem;
 use std::process::exit;
 use std::thread::{self, JoinHandle};
+use std::time::Duration;
 use std::sync::{Arc, Mutex};
 use std::sync::mpsc::channel;
 
 use arguments::Args;
 use execute::pipe::disk::State;
-use tokenizer::{TokenErr, tokenize};
+use input_iterator::{InputIterator, InputsLock};
+use tokenizer::{Token, TokenErr, tokenize};
 
-/// Coercing the `command` `String` into a `&'static str` is required to share it among all threads.
 /// The command string needs to be available in memory for the entirety of the application, so this
-/// is achievable by leaking the memory so it attains a `'static` lifetime.
-unsafe fn leak_command(comm: String) -> &'static str {
-    let static_comm = mem::transmute(&comm as &str);
+/// is achievable by transmuting the lifetime of the reference into a static lifetime. To guarantee
+/// that this is perfectly safe, and that the reference will live outside the scope, the value will
+/// also be leaked so that it is forced to remain in memory for the remainder of the application.
+unsafe fn leak_string(comm: String) -> &'static str {
+    let new_comm = mem::transmute(&comm as &str);
     mem::forget(comm);
-    static_comm
+    new_comm
 }
 
+unsafe fn static_arg(args: &[Token]) -> &'static [Token] {
+    mem::transmute(args)
+}
 
 fn main() {
     // Obtain a handle to standard error's buffer so we can write directly to it.
@@ -56,15 +64,19 @@ fn main() {
     let mut args = Args::new();
     let mut comm = String::with_capacity(128);
     let raw_arguments = env::args().collect::<Vec<String>>();
-    let inputs = init::parse(&mut args, &mut comm, &raw_arguments, &unprocessed_path);
+    args.ninputs = init::parse(&mut args, &mut comm, &raw_arguments, &unprocessed_path);
+
+    // Initialize the `InputIterator` structure, which iterates through all inputs.
+    let inputs = InputIterator::new(&unprocessed_path, args.ninputs)
+        .expect("unable to initialize the InputIterator structure");
 
     // Coerce the `comm` `String` into a `&'static str` so that it may be shared by all threads.
     // This is safe because the original `comm` may no longer be modified due to shadowing rules.
     // It is also safe because `comm` lives to the end of the program.
-    let comm = unsafe { leak_command(comm) };
+    let static_comm = unsafe { leak_string(comm) };
 
     // Attempt to tokenize the command argument into simple primitive placeholders.
-    if let Err(error) = tokenize(&mut args.arguments, comm, &unprocessed_path, args.ninputs) {
+    if let Err(error) = tokenize(&mut args.arguments, static_comm, &unprocessed_path, args.ninputs) {
         let mut stderr = stderr.lock();
         match error {
             TokenErr::File(why) => {
@@ -77,16 +89,13 @@ fn main() {
         exit(1)
     }
 
+    let arguments = unsafe { static_arg(&args.arguments) };
+
     if args.flags & arguments::DRY_RUN != 0 {
-        execute::dry_run(args.flags, inputs, &args.arguments);
+        execute::dry_run(args.flags, inputs, arguments);
     } else {
-        if shell::required(shell::Kind::Tokens(&args.arguments)) {
-            if shell::dash_exists() {
-                args.flags |= arguments::DASH_EXISTS + arguments::SHELL_ENABLED;
-            } else {
-                args.flags |= arguments::SHELL_ENABLED;
-            }
-        }
+        if shell::dash_exists() { args.flags |= arguments::DASH_EXISTS; }
+        if shell::required(shell::Kind::Tokens(arguments)) { args.flags |= arguments::SHELL_ENABLED; }
 
         let shared_input = Arc::new(Mutex::new(inputs));
 
@@ -102,16 +111,25 @@ fn main() {
 
         // The `slot` variable is required by the {%} token.
         if args.flags & arguments::INPUTS_ARE_COMMANDS != 0 {
+            if shell::dash_exists() { args.flags |= arguments::DASH_EXISTS; }
+
             for _ in 0..args.ncores {
                 let flags = args.flags;
-                let exec = execute::ExecInputs {
+
+                let mut exec = execute::ExecInputs {
                     num_inputs: args.ninputs,
-                    delay:      args.delay,
                     timeout:    args.timeout,
-                    memory:     args.memory,
-                    inputs:     shared_input.clone(),
                     output_tx:  output_tx.clone(),
+                    inputs:     InputsLock {
+                        inputs:    shared_input.clone(),
+                        memory:    args.memory,
+                        delay:     args.delay,
+                        has_delay: args.delay != Duration::from_millis(0),
+                        completed: false,
+                        flags:     flags,
+                    }
                 };
+
                 let handle: JoinHandle<()> = thread::spawn(move || {
                     exec.run(flags);
                 });
@@ -120,28 +138,33 @@ fn main() {
                 threads.push(handle);
             }
         } else {
+            shell::set_flags(&mut args.flags, arguments);
+
             for slot in 1..args.ncores+1 {
-                let flags = args.flags;
-                let delay = args.delay;
                 let timeout = args.timeout;
                 let num_inputs = args.ninputs;
-                let memory = args.memory;
-                let inputs = shared_input.clone();
                 let output_tx = output_tx.clone();
-                let arguments = args.arguments.clone();
+                let flags = args.flags;
+
+                let inputs = InputsLock {
+                    inputs:    shared_input.clone(),
+                    memory:    args.memory,
+                    delay:     args.delay,
+                    has_delay: args.delay != Duration::from_millis(0),
+                    completed: false,
+                    flags:     flags,
+                };
 
                 // The command will be built from the arguments, and inputs will be transferred to the command.
                 let handle: JoinHandle<()> = thread::spawn(move || {
-                    let exec = execute::ExecCommands {
+                    let mut exec = execute::ExecCommands {
                         slot:       slot,
                         num_inputs: num_inputs,
                         flags:      flags,
-                        delay:      delay,
                         timeout:    timeout,
-                        memory:     memory,
                         inputs:     inputs,
                         output_tx:  output_tx,
-                        arguments:  &arguments,
+                        arguments:  arguments,
                     };
                     exec.run();
                 });
@@ -151,6 +174,7 @@ fn main() {
             }
         }
 
+        /// Prints messages from executed commands in the correct order.
         execute::receive_messages(input_rx, args, &processed_path, &errors_path);
         for thread in threads { thread.join().unwrap(); }
 
